# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).


from kqcircuits.pya_resolver import pya


def insert_ground_grid(
    target_cell: pya.Cell,
    target_layer: pya.LayerInfo,
    grid_area: pya.Box,
    protection: pya.Region | pya.RecursiveShapeIterator,
    grid_step: int,
    grid_size: int,
    grid_offset: int = 0,
):
    """Generates ground grid as shapes in a target cell, without cell hierarchy.
    This function uses integer database units for all inputs.

    Args:
        target_cell: Cell to place the grid into
        target_layer: Layer to place the grid into
        grid_area: Area to fill with grid
        protection: Region to avoid when filling grid
        grid_step: distance between grid rectangles
        grid_size: size of grid rectangles
        grid_offset: Value between 0 (inclusive) and grid_step/grid_size (exclusive) to place grid rectangle.
            0 (default) for bottom left of grid_step * grid_step tile, increasing integer value places rectangle
            further up and right. Ensures multiple grids don't overlap.
    """
    _, grid_cell = _make_ground_grid_cell(target_layer, grid_area, protection, grid_step, grid_size, grid_offset)

    # Copy shapes from temporary layout to the target cell. This flattens the instances of ``grid_element_cell``.
    cm = pya.CellMapping()
    cm.for_single_cell(target_cell, grid_cell)
    target_cell.copy_tree_shapes(grid_cell, cm)


def make_ground_grid_region(
    grid_area: pya.Box,
    protection: pya.Region | pya.RecursiveShapeIterator,
    grid_step: int,
    grid_size: int,
    grid_offset: int = 0,
) -> pya.Region:
    """Returns ground grid as a ``Region``. This function uses integer database units for all inputs.

    Note: ``insert_ground_grid`` is more efficient if the grid will be inserted into a cell.

    Args:
        target_layer: Layer definition to place the grid into
        grid_area: Area to fill with grid
        protection: Region to avoid when filling grid
        grid_step: distance between grid rectangles
        grid_size: size of grid rectangles
        grid_offset: Value between 0 (inclusive) and grid_step/grid_size (exclusive) to place grid rectangle.
            0 (default) for bottom left of grid_step * grid_step tile, increasing integer value places rectangle
            further up and right. Ensures multiple grids don't overlap.

    Returns: a Region containing the ground grid
    """
    dummy_layer = pya.LayerInfo(1, 0)
    layout, grid_cell = _make_ground_grid_cell(dummy_layer, grid_area, protection, grid_step, grid_size, grid_offset)
    grid_region = pya.Region(grid_cell.begin_shapes_rec(layout.layer(dummy_layer)))
    grid_region.merge()  # Ensure the RecursiveShapeIterator is fully iterated over before we discard ``layout``
    return grid_region


def _make_ground_grid_cell(
    target_layer: pya.LayerInfo,
    grid_area: pya.Box,
    protection: pya.Region | pya.RecursiveShapeIterator,
    grid_step: int,
    grid_size: int,
    grid_offset: int,
) -> tuple[pya.Layout, pya.Cell]:
    """Generates ground grid as cell instances in a new cell in a new layout. The returned ``cell`` contains a child
    cell instance for each grid cell element, and should generally be flattened to avoid having too many cell instances.

    A reference to ``layout`` must be kept as long as ``cell`` is used.

    This function uses integer database units for all inputs.

    Args:
        target_layer: Layer definition to place the grid into
        grid_area: Area to fill with grid
        protection: Region to avoid when filling grid
        grid_step: distance between grid rectangles
        grid_size: size of grid rectangles
        grid_offset: Value between 0 (inclusive) and grid_step/grid_size (exclusive) to place grid rectangle.
            0 (default) for bottom left of grid_step * grid_step tile, increasing integer value places rectangle
            further up and right. Ensures multiple grids don't overlap.

    Returns: tuple ``(layout, cell)`` containing a new ``Layout`` and ``Cell``.
    """
    region_with_ground_grid = pya.Region(grid_area) - protection

    # Create temporary layout for ground grid operations
    layout = pya.Layout()
    layout.insert_layer(target_layer)

    # Create a cell with a single ground grid square
    grid_element_cell = layout.create_cell("grid_element")
    grid_rectangle = pya.Box(
        grid_offset * grid_size,
        grid_offset * grid_size,
        (grid_offset + 1) * grid_size,
        (grid_offset + 1) * grid_size,
    )
    grid_element_cell.shapes(layout.layer(target_layer)).insert(grid_rectangle)

    # Generate the full ground grid as instances of ``grid_element_cell`` in a new cell
    grid_cell = layout.create_cell("grid")
    grid_cell.fill_region(
        region_with_ground_grid,
        grid_element_cell.cell_index(),
        grid_rectangle,
        pya.Vector(grid_step, 0),
        pya.Vector(0, grid_step),
        pya.Point(grid_offset * grid_size, grid_offset * grid_size),
    )

    return layout, grid_cell



# This code is part of KQCircuits
# Copyright (C) 2025 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

import math
from kqcircuits.pya_resolver import pya
from kqcircuits.simulations.epr.util import create_bulk_and_mer_partition_regions


def partition_regions(simulation):
    """Returns list of PartitionRegion objects for Circular Transmon Single Island EPR simulation."""

    regions = []

    # Coupler partition regions - one for each coupler
    for i, (c_angle, c_arc_ampl) in enumerate(zip(simulation.couplers_angle, simulation.couplers_arc_amplitude)):
        # Create pizza slice shape for each coupler
        c_angle_rad = math.radians(float(c_angle))
        c_arc_ampl_rad = math.radians(float(c_arc_ampl))

        # Inner radius (closer to qubit center, with some margin)
        inner_radius = simulation.r_island - 10
        # Outer radius (extends beyond coupler, with extra space)
        outer_radius = simulation.r_island + simulation.ground_gap + 20

        # Angular extent with extra margin
        half_angle = c_arc_ampl_rad / 2 + math.radians(10)  # Extra 10 degrees on each side

        # Create pizza slice polygon without center point to avoid zero width lines
        points = []

        # Add arc points for outer edge
        n_points = 32
        for j in range(n_points + 1):
            angle = c_angle_rad - half_angle + (2 * half_angle * j) / n_points
            x = outer_radius * math.cos(angle)
            y = outer_radius * math.sin(angle)
            points.append(pya.DPoint(x, y) + simulation.refpoints["base"])

        # Add arc points for inner edge (reverse direction)
        for j in range(n_points, -1, -1):
            angle = c_angle_rad - half_angle + (2 * half_angle * j) / n_points
            x = inner_radius * math.cos(angle)
            y = inner_radius * math.sin(angle)
            points.append(pya.DPoint(x, y) + simulation.refpoints["base"])

        coupler_region = pya.DPolygon(points)

        regions += create_bulk_and_mer_partition_regions(
            name=f"{i}cplr",
            face=simulation.face_ids[0],
            metal_edge_dimensions=2.0,
            region=coupler_region,
            vertical_dimensions=3.0,
            bulk=False,
            visualise=True,
        )

    # Junction hall partition region
    squid_angle_rad = math.radians(simulation.squid_angle)
    hall_length = simulation.ground_gap + 10  # Extra space
    hall_width = 8 + 4  # Junction hall width + extra space

    # Create rotated rectangle for junction hall
    hall_rect = pya.DPolygon(
        [
            pya.DPoint(-hall_width / 2, 0),
            pya.DPoint(hall_width / 2, 0),
            pya.DPoint(hall_width / 2, hall_length),
            pya.DPoint(-hall_width / 2, hall_length),
        ]
    )

    # Position at island edge and rotate
    junction_pos = pya.DPoint(
        simulation.r_island * math.cos(squid_angle_rad), simulation.r_island * math.sin(squid_angle_rad)
    )

    hall_transform = pya.DCplxTrans(1, simulation.squid_angle - 90, False, junction_pos + simulation.refpoints["base"])
    hall_region = hall_transform * hall_rect

    regions += create_bulk_and_mer_partition_regions(
        name="leads",
        face=simulation.face_ids[0],
        metal_edge_dimensions=2.0,
        region=hall_region,
        vertical_dimensions=3.0,
        bulk=False,
        visualise=True,
    )

    # Main island complement region
    regions += create_bulk_and_mer_partition_regions(
        name=f"{simulation.face_ids[0]}complement",
        face=simulation.face_ids[0],
        metal_edge_dimensions=2.0,
        region=None,
        vertical_dimensions=3.0,
        bulk=False,
        visualise=True,
    )

    return regions


def correction_cuts(simulation):
    """Returns dictionary of correction cuts for Circular Transmon Single Island EPR simulation."""

    cuts = {}

    # Coupler correction cuts
    for i, (c_angle, c_arc_ampl) in enumerate(zip(simulation.couplers_angle, simulation.couplers_arc_amplitude)):
        c_angle_rad = math.radians(float(c_angle))
        c_arc_ampl_rad = math.radians(float(c_arc_ampl))

        # Cut  outside the hall and cross the coupler slightly on the side
        # mid angle between current cut position and coupler arc amplitude
        half_arc = c_arc_ampl_rad / 2
        cut_angle_offset = half_arc / 2  # Halfway between center and arc edge
        cut_angle = c_angle_rad + cut_angle_offset  # Offset to the side

        # Cut from outside qubit through coupler to island center
        # Start outside the qubit
        outer_point = pya.DPoint(
            (simulation.r_island + simulation.ground_gap + 15) * math.cos(cut_angle),
            (simulation.r_island + simulation.ground_gap + 15) * math.sin(cut_angle),
        )

        # End at island edge
        inner_point = pya.DPoint(
            (simulation.r_island - 10) * math.cos(cut_angle), (simulation.r_island - 10) * math.sin(cut_angle)
        )

        cuts[f"{i}cplrmer"] = {
            "p1": outer_point + simulation.refpoints["base"],
            "p2": inner_point + simulation.refpoints["base"],
        }

    # Junction hall correction cut
    squid_angle_rad = math.radians(simulation.squid_angle)

    # Cut across the width of junction hall
    hall_center = pya.DPoint(
        (simulation.r_island + simulation.ground_gap / 2) * math.cos(squid_angle_rad),
        (simulation.r_island + simulation.ground_gap / 2) * math.sin(squid_angle_rad),
    )

    # Perpendicular direction to junction hall
    perp_angle = squid_angle_rad + math.pi / 2
    cut_length = 15

    p1 = pya.DPoint(
        hall_center.x + cut_length * math.cos(perp_angle), hall_center.y + cut_length * math.sin(perp_angle)
    )
    p2 = pya.DPoint(
        hall_center.x - cut_length * math.cos(perp_angle), hall_center.y - cut_length * math.sin(perp_angle)
    )

    cuts["leadsmer"] = {
        "p1": p1 + simulation.refpoints["base"],
        "p2": p2 + simulation.refpoints["base"],
        "boundary_conditions": {"xmin": {"potential": 0}, "xmax": {"potential": 0}},
    }

    # Main island complement correction cut
    # direction that avoids couplers and junction
    # angle opposite to junction
    complement_angle = math.radians(simulation.squid_angle + 180)

    # Check if this angle conflicts with any coupler
    min_angular_distance = 360
    for c_angle in simulation.couplers_angle:
        angular_dist = abs(float(c_angle) - (simulation.squid_angle + 180))
        if angular_dist > 180:
            angular_dist = 360 - angular_dist
        min_angular_distance = min(min_angular_distance, angular_dist)

    # Close to the coupler
    if min_angular_distance < 30:  # Less than 30 degrees separation
        complement_angle = math.radians(simulation.squid_angle + 150)

    # Cut from island to outside qubit
    # points equidistant from the gap region
    gap_distance = 10  # Same distance from gap as outer point

    inner_point = pya.DPoint(
        (simulation.r_island - gap_distance) * math.cos(complement_angle),
        (simulation.r_island - gap_distance) * math.sin(complement_angle),
    )
    outer_point = pya.DPoint(
        (simulation.r_island + simulation.ground_gap + gap_distance) * math.cos(complement_angle),
        (simulation.r_island + simulation.ground_gap + gap_distance) * math.sin(complement_angle),
    )

    cuts[f"{simulation.face_ids[0]}complementmer"] = {
        "p1": inner_point + simulation.refpoints["base"],
        "p2": outer_point + simulation.refpoints["base"],
    }

    return cuts



# This code is part of KQCircuits
# Copyright (C) 2025 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

import ast
from typing import Tuple, Type

from kqcircuits.pya_resolver import pya
from kqcircuits.elements.element import Element
from kqcircuits.util.library_helper import element_by_class_name

from kqcircuits.elements.airbridges.airbridge import Airbridge
from kqcircuits.elements.airbridge_connection import AirbridgeConnection
from kqcircuits.elements.flip_chip_connectors.flip_chip_connector_rf import FlipChipConnectorRf
from kqcircuits.elements.waveguide_coplanar_splitter import WaveguideCoplanarSplitter

# Speed up lookup for elements most commonly used as Node.element
COMMON_ELEMENTS = {
    cls.__name__: cls for cls in [Airbridge, AirbridgeConnection, FlipChipConnectorRf, WaveguideCoplanarSplitter]
}


class Node:
    """Specifies a single node of a composite waveguide.

    Node is as a ``position`` and optionally other parameters. The ``element`` argument sets an Element
    type that gets inserted in the waveguide. Typically this is an Airbridge, but any element with
    port_a and port_b is supported.

    Args:
        position: The location of the Node. Represented as a DPoint object.
        element: The Element type that gets inserted in the waveguide. None by default.
        inst_name: If an instance name is supplied, the element refpoints will be exposed with that name. Default None.
        align: Tuple with two refpoint names that correspond the input and output point of element, respectively
               Default value (None) uses ``('port_a', 'port_b')``
        angle: Angle of waveguide direction in degrees
        length_before: Length of the waveguide segment before this node
        length_increment: Waveguide length increment produced by meander before this node
        **params: Other optional parameters for the inserted element

    Returns:
        A Node.
    """

    position: pya.DPoint
    element: Type[Element] | None
    inst_name: str | None
    align: Tuple
    angle: float | None
    length_before: float | None
    length_increment: float | None
    meander_direction: int

    def __init__(
        self,
        position: pya.DPoint,
        element: Type[Element] | None = None,
        inst_name: str | None = None,
        align: tuple = tuple(),
        angle: float | None = None,
        length_before: float | None = None,
        length_increment: float | None = None,
        meander_direction: int = 1,
        **params,
    ):
        if isinstance(position, tuple):
            self.position = pya.DPoint(position[0], position[1])
        else:
            self.position = position
        self.element = element
        self.align = align
        self.inst_name = inst_name
        self.angle = angle
        self.length_before = length_before
        self.length_increment = length_increment
        self.meander_direction = meander_direction
        self.params = params

    def __str__(self):
        """
        String representation of a Node, used for serialization and needed for storage in KLayout parameters.

        The corresponding deserialization is implemented in `Node.deserialize`.
        """

        txt = f"{self.position.x}, {self.position.y}"
        if self.element is not None:
            txt += f", '{self.element.__name__}'"

        magic_params = {}
        if self.align:
            magic_params["align"] = self.align
        if self.inst_name:
            magic_params["inst_name"] = self.inst_name
        if self.angle is not None:
            magic_params["angle"] = self.angle
        if self.length_before is not None:
            magic_params["length_before"] = self.length_before
        if self.length_increment is not None:
            magic_params["length_increment"] = self.length_increment
        if self.meander_direction != 1:
            magic_params["meander_direction"] = self.meander_direction

        all_params = {**self.params, **magic_params}
        if all_params:
            for pn, pv in all_params.items():
                if isinstance(pv, pya.DPoint):  # encode DPoint as tuple
                    all_params[pn] = (pv.x, pv.y)
            txt += f", {all_params}"
        return "(" + txt + ")"

    @classmethod
    def deserialize(
        cls,
        serialized_node: (
            tuple[float, float] | tuple[float, float, str | None | dict] | tuple[float, float, str | None, dict]
        ),
    ):
        """
        Create a Node object from a serialized form, such that ``from_serialized(ast.literal_eval(str(node_object)))``
        returns an equivalent copy of ``node_obj``.

        Args:
            serialized_node: serialized node, consisting of a tuple ``(x, y, element_name, params)``, where ``x`` and
            ``y`` are the node coordinates. The string ``element_name`` and dict ``params`` are optional.

        Returns: a Node

        """
        x, y = float(serialized_node[0]), float(serialized_node[1])
        element = None
        params = {}
        if len(serialized_node) > 2:
            if isinstance(serialized_node[2], dict):
                params = serialized_node[2]
            else:
                element = serialized_node[2]
        if len(serialized_node) > 3:
            params = serialized_node[3]

        if element is not None:
            if element in COMMON_ELEMENTS:
                element = COMMON_ELEMENTS[element]
            else:
                element = element_by_class_name(element)

        # re-create DPoint from tuple
        magic_params = ("align", "inst_name", "angle")
        for pn, pv in params.items():
            if isinstance(pv, tuple) and pn not in magic_params:
                if len(pv) < 2:
                    raise ValueError(f"Point parameter {pn} should have two elements")
                params[pn] = pya.DPoint(float(pv[0]), float(pv[1]))

        return cls(pya.DPoint(x, y), element, **params)

    @staticmethod
    def nodes_from_string(nodes: list[str] | str) -> list["Node"]:
        """Converts the human-readable text representation of Nodes to an actual Node object list.

        Needed for storage in KLayout parameters. The string has to conform to a specific format:
        `(x, y, class_str, parameter_dict)`. For example `(0, 500, 'Airbridge', {'n_bridges': 2})`,
        see also the `Node.__str__` method. Empty class_str or parameter_dict may be omitted.

        Returns:
            list of Node objects
        """

        nlas = ", ".join(nodes) if isinstance(nodes, list) else nodes
        node_list = ast.literal_eval(nlas + ",")

        return [Node.deserialize(node) for node in node_list]



# This code is part of KQCircuits
# Copyright (C) 2023 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

from typing import Callable
from kqcircuits.elements.element import Element
from kqcircuits.junctions import junction_type_choices
from kqcircuits.simulations.simulation import Simulation
from kqcircuits.pya_resolver import pya
from kqcircuits.simulations.partition_region import PartitionRegion
from kqcircuits.simulations.port import InternalPort, EdgePort
from kqcircuits.util.parameters import Param, pdt, add_parameters_from
from kqcircuits.util.refpoints import RefpointToInternalPort, RefpointToEdgePort, WaveguideToSimPort, JunctionSimPort


def _get_build_function(
    element_class, ignore_ports, transformation_from_center, sim_junction_type, deembed_cross_sections
):
    def _build_for_element_class(self):
        if sim_junction_type not in junction_type_choices:
            raise ValueError(
                f"Unknown sim_junction_type {sim_junction_type}. "
                f"Junction type should be listed in junction_type_choices"
            )

        simulation_cell = self.add_element(
            element_class, **{**self.get_parameters(), "junction_type": sim_junction_type, "fluxline_type": "none"}
        )

        element_trans = pya.DTrans(0, False, self.box.center())
        if transformation_from_center is not None:
            element_trans *= transformation_from_center(simulation_cell)
        _, refp = self.insert_cell(simulation_cell, element_trans, rec_levels=None)
        self.refpoints = refp

        if deembed_cross_sections is not None:
            deembed_cs_names = deembed_cross_sections.keys()
        else:
            deembed_cs_names = set()

        # Add ports
        port_i = 0
        for port in element_class.get_sim_ports(self):
            if ignore_ports is not None and port.refpoint in ignore_ports:
                continue

            if port.refpoint in deembed_cs_names:
                port.deembed_cross_section = deembed_cross_sections[port.refpoint]

            if isinstance(port, RefpointToInternalPort):
                self.ports.append(
                    InternalPort(
                        number=(port_i := port_i + 1),
                        signal_location=refp[port.refpoint],
                        ground_location=None if not port.ground_refpoint else refp[port.ground_refpoint],
                        resistance=port.resistance,
                        reactance=port.reactance,
                        inductance=port.inductance,
                        capacitance=port.capacitance,
                        face=port.face,
                        junction=port.junction,
                    )
                )
            elif isinstance(port, RefpointToEdgePort):
                self.ports.append(
                    EdgePort(
                        number=(port_i := port_i + 1),
                        signal_location=refp[port.refpoint],
                        resistance=port.resistance,
                        reactance=port.reactance,
                        inductance=port.inductance,
                        capacitance=port.capacitance,
                        deembed_len=port.deembed_len,
                        face=port.face,
                        junction=port.junction,
                        size=port.size,
                        deembed_cross_section=port.deembed_cross_section,
                    )
                )
            elif isinstance(port, WaveguideToSimPort):
                towards = port.towards
                if port.towards is None:
                    towards = f"{port.refpoint}_corner"
                self.produce_waveguide_to_port(
                    refp[port.refpoint],
                    refp[towards],
                    (port_i := port_i + 1),
                    side=port.side,
                    a=port.a,
                    b=port.b,
                    term1=port.term1,
                    turn_radius=port.turn_radius,
                    use_internal_ports=port.use_internal_ports,
                    waveguide_length=port.waveguide_length,
                    face=port.face,
                    airbridge=port.airbridge,
                    deembed_cross_section=port.deembed_cross_section,
                )

            elif isinstance(port, JunctionSimPort):
                self.ports.append(
                    InternalPort(
                        (port_i := port_i + 1),
                        *self.etched_line(refp[port.refpoint], refp[port.other_refpoint]),
                        face=port.face,
                        inductance=self.junction_inductance,
                        capacitance=self.junction_capacitance,
                        junction=True,
                        floating=port.floating,
                    )
                )

    return _build_for_element_class


def get_single_element_sim_class(
    element_class: Element,
    ignore_ports: list[str] | None = None,
    transformation_from_center: Callable[[pya.Cell], pya.DTrans] | None = None,
    partition_region_function: Callable[[Simulation], list[PartitionRegion]] | None = None,
    sim_junction_type: str = "Sim",
    deembed_cross_sections: dict[str] = None,
) -> type[Simulation]:
    """Formulates a simulation class containing a single cell of a given Element class

    Args:
        element_class: an Element class for which a simulation class is returned
        ignore_ports: If list of strings is given, simulation ports will not be created for the given
            refpoints in the simulation class.
        transformation_from_center: If None, simulated element is placed in the middle of simulation's box.
            Otherwise should be a function that takes an element cell as argument and returns a DTrans object.
            The returned transformation is applied to the element cell
            after placing it in the middle of simulation's box.
            The function should not cause any side-effects, i.e. change the cell parameters
        partition_region_function: optional. Function that the simulation instance will use to define
            partition regions, which may look up instances parameters and refpoints to derive the regions.
        deembed_cross_sections: optional dictionary for cross-section simulation that can be used for deembeding.
            The naming convention in the dictionary is `deembed_cross_sections[port_refpoint]=cross_section_name`,
            where `cross_section_name` is the name given to the correction cuts. For example, see
            `simulation.epr.smooth_capacitor.py`, deembed_cross_sections['port_a']='port_amer'.
    """
    overriden_class_attributes = {
        "junction_inductance": Param(pdt.TypeList, "Junction inductance (if junction exists)", 11.497e-9, unit="H"),
        "junction_capacitance": Param(pdt.TypeList, "Junction capacitance (if junction exists)", 0.1e-15, unit="F"),
        "build": _get_build_function(
            element_class, ignore_ports, transformation_from_center, sim_junction_type, deembed_cross_sections
        ),
    }
    if partition_region_function:
        _cache = {}

        def _get_partition_regions(simulation: Simulation):
            if simulation not in _cache:
                _cache[simulation] = partition_region_function(simulation)
            return _cache[simulation]

        overriden_class_attributes["get_partition_regions"] = _get_partition_regions
    element_sim_class = type(
        f"SingleElementSimulationClassFor{element_class.__name__}",
        (Simulation,),
        overriden_class_attributes,
    )
    add_parameters_from(element_class)(element_sim_class)
    return element_sim_class



# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

import math
from typing import Callable
from kqcircuits.pya_resolver import pya
from kqcircuits.simulations.epr.util import (
    extract_child_simulation,
    in_gui,
    EPRTarget,
    get_mer_z,
    create_bulk_and_mer_partition_regions,
    is_flip_chip,
)
from kqcircuits.simulations.partition_region import PartitionRegion
from kqcircuits.simulations.simulation import Simulation
from kqcircuits.util.geometry_helper import arc_points
from kqcircuits.elements.finger_capacitor_square import eval_a2, eval_b2

# Partition region and correction cuts definitions for CircularCapacitor element


def _has_waveguides(simulation):
    """Determines if waveguide regions and cuts are added used"""
    if in_gui(simulation):
        return False
    return simulation.fixed_length > 0 or simulation.waveguide_length > 0 or not simulation.use_internal_ports


def _waveguide_end_dist(simulation):
    """Distance from the center to the end of the waveguide"""
    if simulation.use_internal_ports:
        x_guide = simulation.waveguide_length + simulation.a + 1
        if simulation.fixed_length > 0:
            return x_guide + simulation.fixed_length / 2
        else:
            return x_guide + simulation.r_outer + simulation.ground_gap
    else:
        return simulation.box.width() / 2


def partition_regions(simulation: EPRTarget, prefix: str = "") -> list[PartitionRegion]:

    mer_x_dim = 2.0
    mer_y_dim = 1.0
    # Let's make the mer margin just a bit larger to prevent any artefacts
    region_safety_margin = 1.5  # um
    metal_edge_margin = mer_x_dim + region_safety_margin

    center = simulation.refpoints["base"]

    def _init_part_reg(name, region, mer=True, bulk=True, face=0, override_y_dim=mer_y_dim):
        if in_gui(simulation):
            face_id = simulation.face_ids[face]
        else:
            face_id = simulation.face_stack[face]
        return create_bulk_and_mer_partition_regions(
            name=f"{prefix}{name}",
            face=face_id,
            metal_edge_dimensions=mer_x_dim,
            region=region,
            vertical_dimensions=override_y_dim,
            mer=mer,
            bulk=bulk,
            visualise=True,
        )

    def _symmetric_sector(angle, r_out, angle_in=None, r_in=0, center=center, mirror=False):
        if angle_in is None:
            angle_in = angle

        def _arc(r, angle, mirror):
            half_angle = angle / 2
            start = -half_angle + (math.pi if mirror else 0)
            stop = half_angle + (math.pi if mirror else 0)
            return arc_points(r, start=start, stop=stop, n=simulation.n, origin=center)

        outer_arc = _arc(r_out, angle, mirror)

        if r_in > 0:
            inner_arc = _arc(r_in, angle_in, mirror)
        else:
            inner_arc = [center]

        return pya.DPolygon(inner_arc + list(reversed(outer_arc)))

    def _angle_from_x(p, origin=center):
        return math.atan2(p.y - origin.y, p.x - origin.x)

    r_coupler_out = simulation.r_outer
    r_coupler_in = r_coupler_out - simulation.outer_island_width
    r_tot = r_coupler_out + simulation.ground_gap

    # Find the angles for the inner and outer arc forming the coupler from refpoints
    coupler_p_in = simulation.refpoints["epr_cplr_in"]
    coupler_p_out = simulation.refpoints["epr_cplr_out"]
    angle_in = 2 * _angle_from_x(coupler_p_in)
    angle_out = 2 * _angle_from_x(coupler_p_out)

    inner_gap_region = _symmetric_sector(
        2 * _angle_from_x((coupler_p_in + coupler_p_out) / 2),
        r_coupler_in + simulation.outer_island_width / 2 + region_safety_margin,
    )
    outer_gap_region = _symmetric_sector(
        angle_out, r_tot + 1.5 * metal_edge_margin, r_in=r_coupler_out - simulation.outer_island_width / 2
    )

    # if coupler angle is closer than 45 deg from the x axis, make the region cover the central trace.
    # Otherwise the central trace will be in port_a region
    gap_type = (
        simulation.swept_angle / 2 > 135 and abs(((coupler_p_in + coupler_p_out) / 2).x - center.x) > simulation.r_inner
    )

    # how much the angle needs to be larger to include the MER area within the coupler
    coupler_angle_offset_in = 2 * math.atan(1.5 * metal_edge_margin / (r_coupler_in - metal_edge_margin))
    coupler_angle_offset_out = 2 * math.atan(1.5 * metal_edge_margin / (r_coupler_out + metal_edge_margin))
    if gap_type:
        coupler_gap_region = _symmetric_sector(
            2 * math.pi - angle_out + coupler_angle_offset_out,
            r_coupler_out + metal_edge_margin,
            angle_in=2 * math.pi - angle_in + coupler_angle_offset_in,
            r_in=r_coupler_in - metal_edge_margin,
            mirror=True,
        )
    else:
        coupler_up = pya.DPolygon(
            arc_points(
                r_coupler_out + metal_edge_margin,
                start=angle_out / 2 - coupler_angle_offset_out,
                stop=angle_out / 2 + coupler_angle_offset_out,
                n=simulation.n,
                origin=center,
            )
            + arc_points(
                r_coupler_in - metal_edge_margin,
                start=angle_in / 2 + coupler_angle_offset_in,
                stop=angle_in / 2 - coupler_angle_offset_in,
                n=simulation.n,
                origin=center,
            )
        )
        # mirror the coupler region to the other side
        coupler_down = (
            coupler_up.dup()
            .transform(pya.DTrans(pya.DVector(-center)))
            .transform(pya.DTrans(0, True, pya.DVector(center)))
        )
        coupler_gap_region = [coupler_up, coupler_down]

    lead_region = pya.DBox(
        center + pya.DPoint(-r_tot - metal_edge_margin, -simulation.a / 2 - metal_edge_margin),
        center + pya.DPoint(-simulation.r_inner + 1.5 * metal_edge_margin, simulation.a / 2 + metal_edge_margin),
    )
    a2, b2 = eval_a2(simulation), eval_b2(simulation)
    lead2_region = pya.DBox(
        center + pya.DPoint(r_coupler_out - 1.5 * metal_edge_margin, -a2 / 2 - metal_edge_margin),
        center + pya.DPoint(r_tot + metal_edge_margin, a2 / 2 + metal_edge_margin),
    )

    # Some of the regions overlap so the order of definition is important
    result = []
    result += _init_part_reg("cplr", coupler_gap_region)
    result += _init_part_reg("1lead", lead_region)
    result += _init_part_reg("2lead", lead2_region)

    if _has_waveguides(simulation):
        x_guide = _waveguide_end_dist(simulation)
        # include large bulk area such that all of the waveguide energy can be removed in post-processing
        wg_y_margin = 100
        y1 = simulation.a / 2 + simulation.b + wg_y_margin
        y2 = a2 / 2 + b2 + wg_y_margin

        wg1_region = pya.DBox(
            center + pya.DPoint(-x_guide - metal_edge_margin, -y1),
            center + pya.DPoint(-r_tot - metal_edge_margin / 2, y1),
        )
        wg2_region = pya.DBox(
            center + pya.DPoint(r_tot + metal_edge_margin / 2, -y2),
            center + pya.DPoint(x_guide + metal_edge_margin, y2),
        )

        result += _init_part_reg("port_a", wg1_region, bulk=False)
        result += _init_part_reg("port_b", wg2_region, bulk=False)
        # Making the wg bulk regions very large to contain all wg energy
        result += _init_part_reg("port_a", wg1_region, mer=False, override_y_dim=100)
        result += _init_part_reg("port_b", wg2_region, mer=False, override_y_dim=100)

    result += _init_part_reg("1gap", inner_gap_region)
    result += _init_part_reg("2gap", outer_gap_region)
    result += _init_part_reg(f"{simulation.face_ids[0]}complement", None)

    if is_flip_chip(simulation):
        if simulation.etch_opposite_face:
            result += _init_part_reg(f"{simulation.face_ids[1]}complement", None, bulk=False, face=1)
        result += _init_part_reg(f"{simulation.face_ids[1]}complement", None, mer=False, face=1)

    return result


def correction_cuts(simulation: EPRTarget, prefix: str = "") -> dict[str, dict]:

    # def _init_cut(center, width, edges):
    result = {}

    center = simulation.refpoints["base"]
    # Make cuts this much shorter than allowed by perfect rounding to account for discretization
    cut_length_margin = 0.5

    z_me_b = get_mer_z(simulation, simulation.face_ids[0])
    r_tot = simulation.r_outer + simulation.ground_gap

    a2, b2 = eval_a2(simulation), eval_b2(simulation)

    def _coupler_cut_lim(unit_vec, start_p, origin_p, rlim):
        """
        Find how long a line can be extended from `start_p` towards `unit vec` before being over
        `rlim` away from `origin_p`. Returns the solution with smallest absolute value

        Solves n from || start_p + n*unit_vec - origin_p || =  rlim.

        """
        diff = pya.DVector(start_p) - pya.DVector(origin_p)
        dotp = unit_vec.sprod(diff)
        pm = (dotp**2 - diff.sq_length() + rlim**2) ** 0.5
        return min(abs(dotp + pm), abs(dotp - pm))

    cut_center = center + pya.DPoint((-simulation.r_inner - simulation.r_outer + simulation.outer_island_width) / 2, 0)
    cut_lim = _coupler_cut_lim(
        pya.DVector(0, 1), cut_center, center, simulation.r_outer - simulation.outer_island_width
    )
    half_cut_length = min(cut_lim - cut_length_margin, 30)
    result[f"{prefix}1leadmer"] = {
        "p1": cut_center + pya.DPoint(0, -half_cut_length),
        "p2": cut_center + pya.DPoint(0, half_cut_length),
        "metal_edges": [
            {"x": half_cut_length - simulation.a / 2, "x_reversed": True, "z": z_me_b},
            {"x": half_cut_length + simulation.a / 2, "z": z_me_b},
        ],
        "boundary_conditions": {"xmin": {"potential": 0}, "xmax": {"potential": 0}},
    }

    cut_center = center + pya.DPoint(simulation.r_outer + simulation.ground_gap / 2, 0)
    cut_lim = _coupler_cut_lim(pya.DVector(0, 1), cut_center, center, simulation.r_outer + simulation.ground_gap)
    half_cut_length = min(cut_lim - cut_length_margin, 30)
    result[f"{prefix}2leadmer"] = {
        "p1": cut_center + pya.DPoint(0, -half_cut_length),
        "p2": cut_center + pya.DPoint(0, half_cut_length),
        "metal_edges": [
            {"x": half_cut_length - a2 / 2, "x_reversed": True, "z": z_me_b},
            {"x": half_cut_length + a2 / 2, "z": z_me_b},
        ],
        "boundary_conditions": {"xmin": {"potential": 0}, "xmax": {"potential": 0}},
    }

    if _has_waveguides(simulation):
        x_guide = _waveguide_end_dist(simulation)
        wg_cut_x = (x_guide + r_tot) / 2 - 0.2

        # wg1
        cut_center = center + pya.DPoint(-wg_cut_x, 0)
        half_cut_length = simulation.a / 2 + simulation.b + 20
        result[f"{prefix}port_amer"] = {
            "p1": cut_center + pya.DPoint(0, -half_cut_length),
            "p2": cut_center + pya.DPoint(0, half_cut_length),
            "metal_edges": [
                {"x": half_cut_length - simulation.a / 2 - simulation.b, "z": z_me_b},
                {"x": half_cut_length - simulation.a / 2, "x_reversed": True, "z": z_me_b},
                {"x": half_cut_length + simulation.a / 2, "z": z_me_b},
                {"x": half_cut_length + simulation.a / 2 + simulation.b, "x_reversed": True, "z": z_me_b},
            ],
        }

        # wg2
        cut_center = center + pya.DPoint(wg_cut_x, 0)
        half_cut_length = a2 / 2 + b2 + 20
        result[f"{prefix}port_bmer"] = {
            "p1": cut_center + pya.DPoint(0, -half_cut_length),
            "p2": cut_center + pya.DPoint(0, half_cut_length),
            "metal_edges": [
                {"x": half_cut_length - a2 / 2 - b2, "z": z_me_b},
                {"x": half_cut_length - a2 / 2, "x_reversed": True, "z": z_me_b},
                {"x": half_cut_length + a2 / 2, "z": z_me_b},
                {"x": half_cut_length + a2 / 2 + b2, "x_reversed": True, "z": z_me_b},
            ],
        }

    # 1gap cut
    half_gap = (simulation.r_outer - simulation.outer_island_width - simulation.r_inner) / 2
    cut_center = center + pya.DPoint(simulation.r_inner + half_gap, 0)
    half_cut_length = half_gap + min(20, simulation.outer_island_width - cut_length_margin)
    result[f"{prefix}1gapmer"] = {
        "p1": cut_center + pya.DPoint(-half_cut_length, 0),
        "p2": cut_center + pya.DPoint(half_cut_length, 0),
        "metal_edges": [
            {"x": half_cut_length - half_gap, "z": z_me_b},
            {"x": half_cut_length + half_gap, "x_reversed": True, "z": z_me_b},
        ],
    }

    # 2gap cut
    half_gap = simulation.ground_gap / 2
    # for small swept angle and large a2 we might not be able to take a "gap-type" cross-section so a default one-sided
    # correction is used
    angle_margin = math.radians(0.5)
    wg2_angle = math.atan((b2 + a2 / 2) / (simulation.r_outer + simulation.ground_gap)) + angle_margin
    eff_coupler_angle = math.radians(simulation.swept_angle) / 2 - angle_margin
    cut_angle = (wg2_angle + eff_coupler_angle) / 2
    cut_unit_vector = pya.DVector(math.cos(cut_angle), math.sin(cut_angle))

    if cut_angle > wg2_angle:
        cut_center = center + (simulation.r_outer + half_gap) * cut_unit_vector
        half_cut_length = half_gap + min(simulation.outer_island_width - cut_length_margin, 20)

        result[f"{prefix}2gapmer"] = {
            "p1": cut_center - half_cut_length * cut_unit_vector,
            "p2": cut_center + half_cut_length * cut_unit_vector,
            "metal_edges": [
                {"x": half_cut_length - half_gap, "z": z_me_b},
                {"x": half_cut_length + half_gap, "x_reversed": True, "z": z_me_b},
            ],
        }
    else:
        cut_center = center + simulation.r_outer * cut_unit_vector
        half_cut_length = min(simulation.outer_island_width - cut_length_margin, 30)

        result[f"{prefix}2gapmer"] = {
            "p1": cut_center - half_cut_length * cut_unit_vector,
            "p2": cut_center + half_cut_length * cut_unit_vector,
            "metal_edges": [
                {"x": half_cut_length - half_gap, "z": z_me_b},
            ],
            "boundary_conditions": {"xmax": {"potential": 0}},
        }

    # cplr cut
    cmin = simulation.refpoints["epr_cplr_in"]
    cmax = simulation.refpoints["epr_cplr_out"]

    coupler_edge_center = (cmin + cmax) / 2

    # make the cut across the central trace and both couplers if the remaining angle between coupler and
    # trace is less than 45 degrees and the cut doesn't cause overlaps with inner islands.
    # Otherwise use one-sided cross-section with the 0 potential "at infinity"

    gap_type = simulation.swept_angle / 2 > 135 and abs(coupler_edge_center.x - center.x) > simulation.r_inner
    if gap_type:
        cut_center = pya.DPoint(coupler_edge_center.x, center.y)
        # vertical unit vector
        cut_unit_vector = pya.DVector(0, -1)
        cplr_gap_width = coupler_edge_center.y - center.y - simulation.a / 2

        half_cut_lim_coupler = _coupler_cut_lim(cut_unit_vector, coupler_edge_center, center, simulation.r_outer)

        half_cut_length = cplr_gap_width + simulation.a / 2 + min(half_cut_lim_coupler - cut_length_margin, 20)
        result[f"{prefix}cplrmer"] = {
            "p1": cut_center - half_cut_length * cut_unit_vector,
            "p2": cut_center + half_cut_length * cut_unit_vector,
            "metal_edges": [
                {"x": half_cut_length - simulation.a / 2 - cplr_gap_width, "z": z_me_b},
                {"x": half_cut_length - simulation.a / 2, "x_reversed": True, "z": z_me_b},
                {"x": half_cut_length + simulation.a / 2, "z": z_me_b},
                {"x": half_cut_length + simulation.a / 2 + cplr_gap_width, "x_reversed": True, "z": z_me_b},
            ],
        }
    else:
        cut_center = coupler_edge_center
        # unit vector perpendicaular to the coupler island
        cut_unit_vector = pya.DVector(-(cmax - cmin).y, (cmax - cmin).x)
        cut_unit_vector = (1 / cut_unit_vector.length()) * cut_unit_vector
        # restrict the cut to stay in the coupler island
        half_cut_lim_coupler = _coupler_cut_lim(cut_unit_vector, coupler_edge_center, center, simulation.r_outer)
        # restrict not to cut over the central trace
        half_cut_trace_lim = (
            math.inf
            if cut_unit_vector.y == 0
            else abs((cut_center.y - center.y - simulation.a / 2) / cut_unit_vector.y)
        )
        half_cut_length = min(half_cut_lim_coupler, half_cut_trace_lim)
        # default to max 40
        half_cut_length = min(40, half_cut_length - cut_length_margin)

        result[f"{prefix}cplrmer"] = {
            "p1": cut_center - half_cut_length * cut_unit_vector,
            "p2": cut_center + half_cut_length * cut_unit_vector,
            "metal_edges": [
                {"x": half_cut_length, "z": z_me_b},
            ],
            "boundary_conditions": {"xmax": {"potential": 0}},
        }

    # complement cut on face 0
    # This might contain one or 2 metal edges with maybe non-optimal placement with some parameters
    # However, in the edge cases the corresponding region is almost empty so it doesnt have much effect
    wg1_angle = (
        math.atan(
            (simulation.b + simulation.a / 2 + simulation.etch_opposite_face_margin)
            / (simulation.r_outer + simulation.ground_gap)
        )
        + angle_margin
    )
    cut_unit_vector = pya.DVector(math.cos(math.pi - wg1_angle), math.sin(math.pi - wg1_angle))
    cut_center = center + (simulation.r_outer + simulation.ground_gap) * cut_unit_vector
    half_cut_length = 30
    result[f"{prefix}{simulation.face_ids[0]}complementmer"] = {
        "p1": cut_center + half_cut_length * cut_unit_vector,
        "p2": cut_center - half_cut_length * cut_unit_vector,
        "metal_edges": [
            {"x": half_cut_length, "z": z_me_b},
        ],
        "boundary_conditions": {"xmax": {"potential": 1}},
    }

    if is_flip_chip(simulation) and simulation.etch_opposite_face:
        cut_center = (
            center
            + (simulation.r_outer + simulation.ground_gap + simulation.etch_opposite_face_margin) * cut_unit_vector
        )
        half_cut_length = 30
        z_me_t = get_mer_z(simulation, simulation.face_ids[1])
        result[f"{prefix}{simulation.face_ids[1]}complementmer"] = {
            "p1": cut_center + half_cut_length * cut_unit_vector,
            "p2": cut_center - half_cut_length * cut_unit_vector,
            "metal_edges": [
                {"x": half_cut_length, "z": z_me_t},
            ],
            "boundary_conditions": {"xmax": {"potential": 1}},
        }

    return result


def extract_circular_capacitor_from(
    simulation: EPRTarget, refpoint_prefix: str, parameter_remap_function: Callable[[EPRTarget, str], any]
) -> Simulation:
    # fmt: off
    return extract_child_simulation(
        simulation,
        refpoint_prefix,
        parameter_remap_function,
        [
            "n", "a", "a2", "b", "b2", "face_stack", "waveguide_length",
            "fixed_length", "r_inner", "r_outer", "swept_angle", "outer_island_width", "ground_gap",
            "etch_opposite_face", "etch_opposite_face_margin", "use_internal_ports", "box",
        ],
    )
    # fmt: on



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).
import os
from kqcircuits.defaults import node_editor_valid_elements
from kqcircuits.pya_resolver import pya, is_standalone_session
from kqcircuits.util.gui_helper import node_from_text, get_nodes_near_position, replace_node, node_to_text


if is_standalone_session():
    # Can't define plugins in standalone KLayout library. Just define stubs
    class EditNodePlugin:
        pass

    class EditNodePluginFactory:
        pass

else:

    class EditNodePlugin(pya.Plugin):
        def __init__(self, manager, view):
            self.is_active = False
            self.manager = manager
            self.view = view
            self.selection = None
            self.last_dialog_position = None
            self.capture_range = 10  # Mouse capture and marker size in pixels
            self.last_mouse_position = pya.DPoint(0, 0)
            self.create_dialog()

        def create_dialog(self):
            dialog = pya.QDialog(pya.Application.instance().main_window())
            self.dialog = dialog
            dialog.windowTitle = "Edit Node"
            dialog.setModal(False)

            form = pya.QFormLayout(dialog)

            def add_row(label, widget):
                form.addRow(label, widget)
                return widget

            self.text_x = add_row("x position (µm): ", pya.QLineEdit("0.0", dialog))
            self.text_y = add_row("y position (µm):", pya.QLineEdit("0.0", dialog))

            form.addRow(pya.QLabel("Optional arguments, leave empty for default:", dialog))

            self.text_inst_name = add_row("Instance name:", pya.QLineEdit("", dialog))
            self.text_angle = add_row("Angle (degrees):", pya.QLineEdit("", dialog))
            self.text_length_before = add_row("Length before (µm):", pya.QLineEdit("", dialog))
            self.text_length_increment = add_row("Length increment (µm):", pya.QLineEdit("", dialog))
            self.text_element = add_row("Element:", pya.QComboBox(dialog))
            self.text_element.addItems([""] + node_editor_valid_elements)
            self.text_element.editable = True
            self.text_align = add_row("Refpoint names to align to (input,output):", pya.QLineEdit("", dialog))
            self.text_params = add_row("Element parameters:", pya.QPlainTextEdit("", dialog))
            self.button_apply = add_row("", pya.QPushButton("Apply", dialog))
            self.button_apply.clicked(self.update_node_from_form)

        def update_node_from_form(self):
            if self.selection is None:
                return

            try:
                new_node = node_from_text(
                    self.text_x.text,
                    self.text_y.text,
                    self.text_element.currentText,
                    self.text_inst_name.text,
                    self.text_angle.text,
                    self.text_length_before.text,
                    self.text_length_increment.text,
                    self.text_align.text,
                    self.text_params.toPlainText(),
                )
            except ValueError as e:
                pya.MessageBox.warning(type(e).__name__, str(e), pya.MessageBox.Ok)
                return

            self.manager.transaction("Edit node")
            replace_node(self.selection["instance"], self.selection["node_index"], new_node)
            self.manager.commit()
            self.selection["node"] = new_node
            self.update()

        def update_form_from_node(self, node):
            (
                self.text_x.text,
                self.text_y.text,
                self.text_element.currentText,
                self.text_inst_name.text,
                self.text_angle.text,
                self.text_length_before.text,
                self.text_length_increment.text,
                self.text_align.text,
                self.text_params.plainText,
            ) = node_to_text(node)

        def deselect(self):
            self.selection = None
            self.last_dialog_position = (self.dialog.x, self.dialog.y)
            self.dialog.hide()

        def select(self, instance, node_index, node):
            if self.selection is not None:
                self.deselect()

            marker = pya.Marker(self.view)
            position = instance.dcplx_trans * node.position
            size = pya.DVector(self.capture_range, self.capture_range) / self.view.viewport_trans().mag
            marker.set(pya.DBox(position - size, position + size))
            marker.vertex_size = 0
            marker.line_style = 2

            self.update_form_from_node(node)
            if self.last_dialog_position is not None:
                self.dialog.move(*self.last_dialog_position)
            self.dialog.show()

            self.selection = {
                "instance": instance,
                "instance_trans": instance.dcplx_trans,
                "node_index": node_index,
                "node": node,
                "marker": marker,
            }

        def activated(self):
            self.is_active = True

        def deactivated(self):
            self.deselect()
            self.is_active = False

        def mouse_click_event(self, p, buttons, prio):
            if prio and buttons == pya.ButtonState.LeftButton and self.is_active:
                cell_view = self.view.active_cellview()
                node_data = get_nodes_near_position(
                    cell_view.cell, p, self.capture_range / self.view.viewport_trans().mag
                )
                if len(node_data) == 1:
                    wg_inst, node, node_index = node_data[0]
                    self.select(wg_inst, node_index, node)
                else:
                    self.deselect()
                return True
            else:
                return False

        def mouse_moved_event(self, p, _, __):
            # Store current mouse position for use in tracking_position
            self.last_mouse_position = p

        def has_tracking_position(self):
            # We provide a custom tracking position only if a node is selected
            return self.selection is not None

        def tracking_position(self):
            if self.selection is not None:
                position = self.selection["instance_trans"].inverted() * self.last_mouse_position
                return position
            else:
                return pya.DPoint(0, 0)  # Only reached if something in the internal state is broken

        def update(self):
            # Update marker size to stay constant in pixels
            if self.selection is not None:
                marker = self.selection["marker"]
                position = self.selection["instance_trans"] * self.selection["node"].position
                size = pya.DVector(self.capture_range, self.capture_range) / self.view.viewport_trans().mag
                marker.set(pya.DBox(position - size, position + size))

    class EditNodePluginFactory(pya.PluginFactory):
        def __init__(self):
            if pya.Application.instance().is_editable():
                icon_path = os.path.join(os.path.dirname(__file__), "edit_node_plugin.png")
                self.register(1000, "kqc_edit_node", "Edit Node", icon_path)

                # Set tooltip to something more helpful
                toolbar_action = pya.MainWindow.instance().menu().action("@toolbar.kqc_edit_node")
                if toolbar_action:
                    toolbar_action.tool_tip = "Edit individual Node properties of WaveguideComposite elements"

        def create_plugin(self, manager, _, view):
            return EditNodePlugin(manager, view)



# This code is part of KQCircuits
# Copyright (C) 2025 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).


from kqcircuits.pya_resolver import pya
from kqcircuits.simulations.export.cross_section.cut_simulation import CutSimulation
from kqcircuits.simulations.simulation import Simulation


def create_cross_sections_from_simulations(
    simulations: list[Simulation],
    cuts: tuple[pya.DPoint, pya.DPoint] | list[tuple[pya.DPoint, pya.DPoint]],
    ma_permittivity: float = 0,
    ms_permittivity: float = 0,
    sa_permittivity: float = 0,
    ma_thickness: float = 0,
    ms_thickness: float = 0,
    sa_thickness: float = 0,
    vertical_cull: tuple[float, float] | None = None,
    mer_box: pya.DBox | list[pya.DBox] | None = None,
    magnification_order: int = 0,
    layout: pya.Layout | None = None,
    sim_names: list[str] | None = None,
    sim_class: type[CutSimulation] = CutSimulation,
    **kwargs,
) -> list[CutSimulation]:
    """Create cross-sections of all simulation geometries in the list.
    Will set 'box' and 'cell' parameters according to the produced cross-section geometry data.

    Args:
        simulations: List of Simulation objects, usually produced by a sweep
        cuts: 1. A tuple (p1, p2), where p1 and p2 are endpoints of a cross-section cut or
              2. a list of such tuples such that each Simulation object gets an individual cut
        ma_permittivity: Permittivity of metal–vacuum (air) interface
        ms_permittivity: Permittivity of metal–substrate interface
        sa_permittivity: Permittivity of substrate–vacuum (air) interface
        ma_thickness: Thickness of metal–vacuum (air) interface
        ms_thickness: Thickness of metal–substrate interface
        sa_thickness: Thickness of substrate–vacuum (air) interface
        vertical_cull: Tuple of two y-coordinates, will cull all geometry not in-between the y-coordinates.
            None by default, which means all geometry is retained.
        mer_box: If set as pya.DBox, will create a specified box as metal edge region,
            meaning that the geometry inside the region are separated into different layers with '_mer' suffix
        magnification_order: Increase magnification of simulation geometry to accomodate more precise spacial units.
            0 =   no magnification with 1e-3 dbu
            1 =  10x magnification with 1e-4 dbu
            2 = 100x magnification with 1e-5 dbu etc
            Consider setting non-zero value when using oxide layers with < 1e-3 layer thickness
        layout: predefined layout for the cross-section simulation. If not set, will create new layout.
        sim_names: Names for the created cross-section simulations. If not given, names of parent simulations are used
        sim_class: CutSimulation or its subclass used for processing the cross-section layers
        kwargs: Additional arguments passed to sim_class

    Returns:
        List of XSectionSimulation objects for each Simulation object in simulations
    """
    if isinstance(cuts, tuple):
        cuts = [cuts] * len(simulations)
    cuts = [tuple(c if isinstance(c, pya.DPoint) else c.to_p() for c in cut) for cut in cuts]
    if len(simulations) != len(cuts):
        if len(simulations) != 1:
            raise ValueError("Number of cuts did not match the number of simulations")
        simulations = simulations * len(cuts)

    if not layout:
        layout = pya.Layout()

    sim_names = (sim_names or []) + [None] * (len(cuts) - len(sim_names or []))

    # Increase database unit accuracy in layout if bigger magnification_order set
    if magnification_order > 0:
        layout.dbu = 10 ** (-3 - magnification_order)

    # Collect cross section simulation sweeps
    return [
        sim_class(
            layout,
            name=(name or simulation.name),
            source_sim=simulation,
            cut_start=cut[0],
            cut_end=cut[1],
            cut_bottom=-1e30 if vertical_cull is None else min(vertical_cull),
            cut_top=1e30 if vertical_cull is None else max(vertical_cull),
            tls_layer_thickness=[ma_thickness, ms_thickness, sa_thickness],
            tls_layer_material=["ma", "ms", "sa"],
            material_dict={
                "ma": {"permittivity": ma_permittivity},
                "ms": {"permittivity": ms_permittivity},
                "sa": {"permittivity": sa_permittivity},
            },
            region_map={} if mer_box is None else {"_mer": mer_box if isinstance(mer_box, list) else [mer_box]},
            **kwargs,
        )
        for simulation, cut, name in zip(simulations, cuts, sim_names)
    ]


def visualise_cross_section_cut_on_original_layout(
    simulations: list[Simulation],
    cuts: tuple[pya.DPoint, pya.DPoint] | list[tuple[pya.DPoint, pya.DPoint]],
    cut_label: str = "cut",
    width_ratio: float = 0.0,
) -> None:
    """Visualise requested cross section cuts on the original simulation layout.

    Will add a rectangle between two points of the cut, and two text points into layer "cross_section_cut"::

        * f"{cut_label}_1" representing the left side of the cross section simulation
        * f"{cut_label}_2" representing the right side of the cross section simulation

    In case the export takes cross sections for one simulation multiple times, this function
    can be called on same simulation sweep multiple times so that multiple cuts can be visualised
    in the same layout. In such case it is recommended to differentiate the cuts using `cut_label`.

    Args:
        simulations: list of simulations from which cross sections are taken. After this call these simulations
            will be modified to include the visualised cuts.
        cuts: 1. A tuple (p1, p2), where p1 and p2 are endpoints of a cross-section cut or
              2. a list of such tuples such that each Simulation object gets an individual cut
        cut_label: prefix of the two text points shown for the cut
        width_ratio: rectangles visualising cuts will have a width of length of the cut multiplied by width_ratio.
            By default will set 0 width line, which is visualised in KLayout.
    """
    if isinstance(cuts, tuple):
        cuts = [cuts] * len(simulations)
    cuts = [tuple(c if isinstance(c, pya.DPoint) else c.to_p() for c in cut) for cut in cuts]
    if len(simulations) != len(cuts):
        if len(simulations) != 1:
            raise ValueError("Number of cuts did not match the number of simulations")
        simulations = simulations * len(cuts)
    for simulation, cut in zip(simulations, cuts):
        cut_length = (cut[1] - cut[0]).length()
        marker_path = pya.DPath(cut, cut_length * width_ratio).to_itype(simulation.layout.dbu)
        # Prevent .OAS saving errors by rounding integer value of path width to even value
        marker_path.width -= marker_path.width % 2
        marker = pya.Region(marker_path)
        simulation.visualise_region(marker, cut_label, "cross_section_cut", cut)



# This code is part of KQCircuits
# Copyright (C) 2025 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

from typing import List
from dataclasses import dataclass

from kqcircuits.pya_resolver import pya


@dataclass
class InstanceHierarchy:
    """Data structure holding the instance hierarchy of a single cell instance"""

    instance: pya.Instance
    trans: pya.DCplxTrans
    parent_instances: List[pya.Instance]
    top_cell: pya.Cell


def get_cell_instance_hierarchy(layout: pya.Layout, cell_index: int) -> List[InstanceHierarchy]:
    """Find all instances of a cell and their transforms (DCplxTrans) in the global coordinate system.
    Resolves the full cell hierarchy.

    Args:
      layout: Layout object
      cell_index: Cell index of the cell to find instances of

    Returns: list of ``InstanceHierarchy`` structrures describing the cell hierarchy of each instance
    """

    to_do = [(cell_index, pya.DCplxTrans(), None, [])]
    results = []

    # Traverse the cell hierarchy upwards from the current cell, and accumulate the transformations at each hierarchy
    # level until we reach the top cell, such that we have the global transform of each instance.
    while len(to_do) > 0:
        current_cell_index, input_trans, input_inst, parent_instances = to_do.pop(0)
        cell = layout.cell(current_cell_index)
        has_parents = False
        for parent_inst in cell.each_parent_inst():
            inst = parent_inst.child_inst()
            if input_inst is None:
                input_inst = inst
            else:
                parent_instances = parent_instances + [inst]
            trans = inst.dcplx_trans
            to_do.append((parent_inst.parent_cell_index(), trans * input_trans, input_inst, parent_instances))
            has_parents = True
        if not has_parents and input_inst is not None:
            results.append(
                InstanceHierarchy(
                    instance=input_inst, trans=input_trans, parent_instances=parent_instances, top_cell=cell
                )
            )
    return results


def formatted_cell_instance_hierarchy(inst_data: InstanceHierarchy) -> str:
    """Create formatted list showing the instance hierarchy of all instances of a cell, including the global
    transformation of the instance and the instance names along the hierarchy if they are defined.

    Args:
         layout: Layout object
         cell_index: Cell index of the cell to find instances of

    Returns: formatted string (multiple lines)
    """

    def _format_inst(inst, trans=None):
        inst_name = inst.property("id")
        r = f"{inst.cell.name}"
        if inst_name is not None:
            r += f' "{inst_name}"'
        if trans is not None:
            r += f" at {str(trans)}"
        return r

    def _spacing(i):
        return " " * 2 * (i + 1) + "-> "

    res = _spacing(-1) + f"{inst_data.top_cell.name}\n"
    for i, child in enumerate(reversed(inst_data.parent_instances)):
        res += _spacing(i) + _format_inst(child) + "\n"
    res += _spacing(len(inst_data.parent_instances)) + _format_inst(inst_data.instance, inst_data.trans) + "\n"
    return res


def transform_top_cell(
    transform: pya.DCplxTrans | pya.DTrans, layout: pya.Layout, cell: None | pya.Cell = None
) -> None:
    """Apply transformation to a top cell. Useful for chips.

    An alternative to layout.transform, but it preserves transformation hierarchy of instances.
    KLayout's layout.transform function tends to make transformation displacement to be arbitrary,
    and compensate for it by shifting polygons within instances.
    ``transform_top_cell`` will make sure that polygons won't shift in relation to its instance's
    origin within instance hierarchy.

    Args:
        transform: Transformation to apply
        layout: Layout where the top cell is. Will contain changed geometry after this call.
        cell: Optional, to specify which top cell to transform within layout.
    """
    # Just pick first top cell in layout if not specified
    if cell is None:
        cell = layout.top_cells()[0]
    # Transform immediate child instances
    for i in cell.each_inst():
        i.transform(transform)
    # Transform polygons owned by the top cell. In KQC those tend to be ground grid etc.
    for layer in layout.layer_infos():
        cell.shapes(layout.layer(layer)).transform(transform)



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).


import math
from kqcircuits.pya_resolver import pya
from kqcircuits.util.label import produce_label, LabelOrigin
from kqcircuits.util.parameters import Param, pdt
from kqcircuits.elements.element import Element
from kqcircuits.elements.markers.marker import Marker
from kqcircuits.defaults import default_brand, default_marker_type


class ChipFrame(Element):
    """The PCell declaration for a chip frame.

    The chip frame consists of a dicing edge, and labels and markers in the corners.

     .. MARKERS_FOR_PNG 0,5000,10000,5000 5000,0,5000,10000
    """

    box = Param(
        pdt.TypeShape,
        "Border",
        pya.DBox(pya.DPoint(0, 0), pya.DPoint(10000, 10000)),
        docstring="Bounding box of the chip frame",
    )
    dice_width = Param(pdt.TypeDouble, "Dicing width", 200, unit="μm")
    dice_grid_margin = Param(
        pdt.TypeDouble,
        "Margin between dicing edge and ground grid",
        100,
        docstring="Margin of the ground grid avoidance layer for dicing edge",
    )
    name_mask = Param(pdt.TypeString, "Name of the mask", "M000")  # string '_3' will leave empty space for M000
    name_chip = Param(pdt.TypeString, "Name of the chip", "CTest")
    name_copy = Param(pdt.TypeString, "Name of the copy", None)
    name_brand = Param(pdt.TypeString, "Name of the brand", default_brand)
    text_margin = Param(
        pdt.TypeDouble, "Margin for labels", 100, docstring="Margin of the ground grid avoidance layer around the text"
    )
    marker_dist = Param(
        pdt.TypeDouble,
        "Marker distance from edges",
        1500,
        docstring="Distance of markers from closest edges of the chip face",
    )
    diagonal_squares = Param(pdt.TypeInt, "Number of diagonal squares for the markers", 10)
    marker_types = Param(
        pdt.TypeList,
        "Marker type for each chip corner, clockwise starting from lower left",
        default=[default_marker_type] * 4,
    )
    chip_dicing_width = Param(pdt.TypeDouble, "Width of the chip dicing reference line", 10.0, unit="µm")
    chip_dicing_line_length = Param(pdt.TypeDouble, "Length of the chip dicing reference line", 100.0, unit="µm")
    chip_dicing_gap_length = Param(pdt.TypeDouble, "Gap between two chip dicing reference dashes", 50.0, unit="µm")
    chip_dicing_in_base_metal = Param(pdt.TypeBoolean, "Insert chip dicing lines in base metal addition", False)
    face_label_show = Param(pdt.TypeBoolean, "Show face label below variant label", True)
    face_label_size = Param(pdt.TypeDouble, "Vertical size of character in face label", 75, unit="µm")
    face_label_margin = Param(
        pdt.TypeDouble,
        "Margin for face label, fills ground grid avoidance around label",
        50,
        docstring="Margin of the ground grid avoidance layer around the text in face label",
    )
    face_label_x = Param(pdt.TypeDouble, "X offset of face label, from top right corner of chip", 100, unit="µm")
    face_label_y = Param(pdt.TypeDouble, "Y offset of face label, from top right corner of chip", 450, unit="µm")

    def build(self):
        """Produces dicing edge, markers, labels and ground grid for the chip face."""
        self._produce_dicing_edge()
        self._produce_labels()
        self._produce_markers()

    def _produce_labels(self):
        x_min, x_max, y_min, y_max = self._box_points()
        chip_name = self.name_chip
        labels = [self.name_mask, chip_name, self.name_copy, self.name_brand]
        self._produce_label(labels[0], pya.DPoint(x_min, y_max), LabelOrigin.TOPLEFT)
        if self.name_chip:
            self._produce_label(labels[1], pya.DPoint(x_max, y_max), LabelOrigin.TOPRIGHT)
        if self.face_label_show:
            self._produce_face_label()
        self._produce_label(labels[2], pya.DPoint(x_max, y_min), LabelOrigin.BOTTOMRIGHT)
        self._produce_label(labels[3], pya.DPoint(x_min, y_min), LabelOrigin.BOTTOMLEFT)

    def _produce_face_label(self):
        """Face label ('1t1', '2b1', etc) placed below the variant label
        has separately configured dimensions.
        """
        _, x_max, _, y_max = self._box_points()
        self._produce_label(
            self.face()["id"],
            pya.DPoint(x_max - self.face_label_x, y_max - self.face_label_y),
            LabelOrigin.TOPRIGHT,
            size=self.face_label_size,
            margin=self.face_label_margin,
        )

    def _produce_label(self, label, location, origin, size=None, margin=None):
        """Produces Text PCells with text `label` with `origin` of the text at `location`.

        Wrapper for the stand alone function `produce_label`.
        If size argument not set, will scale text size with chip dimension for chips smaller than 7 mm.

        Args:
            label: the produced text
            location: DPoint of the location of the text
            origin: LabelOrigin of the corner of the label to be placed at the location
            size: Vertical size of label character
            margin: Margin width within label region

        Effect:
            label PCells added to the layout into the parent PCell

        Returns:
            pya.DBox of the extents of produced label, positioned at chip coordinates.
        """
        if margin is None:
            margin = self.text_margin
        if size is None:
            size = 350 * min(1, self.box.width() / 7000, self.box.height() / 7000)
        return produce_label(
            self.cell,
            label,
            location,
            origin,
            self.dice_width,
            margin,
            [self.face()["base_metal_gap_wo_grid"], self.face()["base_metal_gap_for_EBL"]],
            self.face()["ground_grid_avoidance"],
            size,
        )

    def _produce_markers(self):
        x_min, x_max, y_min, y_max = self._box_points()
        if len(self.marker_types) == 4:
            self._produce_marker(
                self.marker_types[0],
                pya.DTrans(x_min + self.marker_dist, y_min + self.marker_dist) * pya.DTrans.R180,
                self.face()["id"] + "_marker_sw",
            )
            self._produce_marker(
                self.marker_types[3],
                pya.DTrans(x_max - self.marker_dist, y_min + self.marker_dist) * pya.DTrans.R270,
                self.face()["id"] + "_marker_se",
            )
            self._produce_marker(
                self.marker_types[1],
                pya.DTrans(x_min + self.marker_dist, y_max - self.marker_dist) * pya.DTrans.R90,
                self.face()["id"] + "_marker_nw",
            )
            self._produce_marker(
                self.marker_types[2],
                pya.DTrans(x_max - self.marker_dist, y_max - self.marker_dist) * pya.DTrans.R0,
                self.face()["id"] + "_marker_ne",
            )
        else:
            print("Warning: chip frame markers need to be for all four corners")

    def _produce_marker(self, marker_type, trans, name):
        if not marker_type:
            return
        cell_marker = self.add_element(Marker, marker_type=marker_type)
        self.insert_cell(cell_marker, trans)
        self.refpoints[name] = trans.disp

    def _produce_dicing_edge(self):
        shape = pya.DPolygon(self._border_points(self.dice_width))
        self.cell.shapes(self.get_layer("base_metal_gap_wo_grid")).insert(shape)
        self.cell.shapes(self.get_layer("base_metal_gap_for_EBL")).insert(shape)

        protection = pya.DPolygon(self._border_points(self.dice_width + self.dice_grid_margin, self.margin))
        self.cell.shapes(self.get_layer("ground_grid_avoidance")).insert(protection)

        box_points = self._box_points()
        p1 = pya.DPoint(box_points[0], box_points[2])
        p2 = pya.DPoint(box_points[1], box_points[2])
        p3 = pya.DPoint(box_points[0], box_points[3])
        p4 = pya.DPoint(box_points[1], box_points[3])
        self._produce_lines_along_edge(p1.y, p3.y, True, p1)
        self._produce_lines_along_edge(p1.x, p2.x, False, p1)
        self._produce_lines_along_edge(p2.y, p4.y, True, p2)
        self._produce_lines_along_edge(p3.x, p4.x, False, p3)

    def _box_points(self):
        """Returns x_min, x_max, y_min, y_max for the given box."""
        x_min = min(self.box.p1.x, self.box.p2.x)
        x_max = max(self.box.p1.x, self.box.p2.x)
        y_min = min(self.box.p1.y, self.box.p2.y)
        y_max = max(self.box.p1.y, self.box.p2.y)
        return x_min, x_max, y_min, y_max

    def _border_points(self, w, extension=0):
        """Returns a set of points forming frame with outer edge on the chip boundaries, and frame thickness ``w``.
        Optional parameter ``extension`` extends the outer edge of the box.
        """
        x_min, x_max, y_min, y_max = self._box_points()
        points = [
            pya.DPoint(x_min - extension, y_min - extension),
            pya.DPoint(x_max + extension, y_min - extension),
            pya.DPoint(x_max + extension, y_max + extension),
            pya.DPoint(x_min - extension, y_max + extension),
            pya.DPoint(x_min - extension, y_min - extension),
            pya.DPoint(x_min + w, y_min + w),
            pya.DPoint(x_min + w, y_max - w),
            pya.DPoint(x_max - w, y_max - w),
            pya.DPoint(x_max - w, y_min + w),
            pya.DPoint(x_min + w, y_min + w),
        ]
        return points

    def _produce_lines_along_edge(self, line_start, line_end, is_vertical, position):
        """Adds chip dicing reference lines along one edge of the chip.
        The line pattern is tied to the global coordinate system

        Args:
            line_start: One-dimensional coordinate of start of the line
            line_end: One-dimensional coordinate of end of the line
            is_vertical: True to draw a line along y-axis, False to draw along x-axis
            position: A DPoint that is in the line
        """
        start = line_start
        segment_length = self.chip_dicing_line_length + self.chip_dicing_gap_length
        end = math.floor(line_start / segment_length) * segment_length + self.chip_dicing_line_length
        if end > start:
            self._add_chip_dicing_line_dash(start, end, is_vertical, position)
        while True:
            start = end + self.chip_dicing_gap_length
            if start > line_end:
                break
            end = start + self.chip_dicing_line_length
            if end > line_end:
                self._add_chip_dicing_line_dash(start, line_end, is_vertical, position)
                break
            self._add_chip_dicing_line_dash(start, end, is_vertical, position)

    def _add_chip_dicing_line_dash(self, start, end, is_vertical, position):
        """Adds a dash as part of a chip dicing reference line.

        Args:
            start: One-dimensional coordinate of start of the dash
            end: One-dimensional coordinate of end of the dash
            is_vertical: True to draw a dash along y-axis, False to draw along x-axis
            position: A DPoint that is in the line to which the dash belongs to
        """
        if is_vertical:
            box = pya.DBox(position.x - self.chip_dicing_width / 2, start, position.x + self.chip_dicing_width / 2, end)
        else:
            box = pya.DBox(start, position.y - self.chip_dicing_width / 2, end, position.y + self.chip_dicing_width / 2)
        self.cell.shapes(self.get_layer("chip_dicing")).insert(box)
        if self.chip_dicing_in_base_metal:
            self.cell.shapes(self.get_layer("base_metal_addition")).insert(box)



# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).
import logging

from kqcircuits.simulations.simulation import Simulation
from kqcircuits.simulations.cross_section_simulation import CrossSectionSimulation
from kqcircuits.simulations.export.ansys.ansys_solution import (
    AnsysEigenmodeSolution,
    AnsysCurrentSolution,
    AnsysVoltageSolution,
    AnsysHfssSolution,
    AnsysCrossSectionSolution,
    AnsysSolution,
)
from kqcircuits.simulations.export.elmer.elmer_solution import (
    ElmerVectorHelmholtzSolution,
    ElmerCapacitanceSolution,
    ElmerCrossSectionSolution,
    ElmerEPR3DSolution,
)


def validate_simulation(simulation, solution):
    """Analyses Simulation and Solution objects and raises an error if inconsistencies in configuration are found.

    Args:
        simulation: A Simulation object.
        solution: A Solution object.
    Raises:
        Errors when validation criteria are not met.
    """
    simulation_and_solution_types_match(simulation, solution)
    flux_integration_layer_exists_if_needed(simulation, solution)
    london_penetration_depth_with_ansys(simulation, solution)

    # Run these checks only for 3D simulations
    if isinstance(simulation, Simulation):
        has_no_ports_when_required(simulation, solution)
        has_edgeport_when_forbidden(simulation, solution)
        check_partition_region_naming(simulation, solution)
        check_tls_sheet_generation(simulation)
        check_tls_sheets_by_solution(simulation, solution)


def simulation_and_solution_types_match(simulation, solution):
    """Validation check: ensures that a simulation and solution types match.
    Args:
        simulation: A Simulation object.
        solution: A Solution object.
    Raises:
        Errors when validation criteria are not met.
    """
    if isinstance(simulation, CrossSectionSimulation) != isinstance(
        solution, (ElmerCrossSectionSolution, AnsysCrossSectionSolution)
    ):
        raise ValidateSimError(
            f"Simulation '{simulation.name}' is incompatible with {type(solution)}",
            validation_type=simulation_and_solution_types_match.__name__,
        )


def has_no_ports_when_required(simulation, solution):
    """Validation check: ensures that a simulation object has ports when the solution type requires it.
    Args:
        simulation: A Simulation object.
        solution: A Solution object.
    Raises:
        Errors when validation criteria are not met.
    """
    port_names = get_port_names(simulation)
    sim_name = simulation.name
    if not port_names and isinstance(
        solution,
        (
            AnsysHfssSolution,
            AnsysVoltageSolution,
            AnsysCurrentSolution,
            ElmerVectorHelmholtzSolution,
            ElmerCapacitanceSolution,
        ),
    ):
        raise ValidateSimError(
            f"Simulation '{sim_name}' has no ports assigned. This is incompatible with {type(solution)}",
            validation_type=has_no_ports_when_required.__name__,
        )


def has_edgeport_when_forbidden(simulation, solution):
    """Validation check: ensure that if at least one "EdgePort" is present, some solution types can't be chosen.
    Args:
        simulation: A Simulation object.
        solution: A Solution object.
    Raises:
        Errors when validation criteria are not met.
    """
    port_names = get_port_names(simulation)
    sim_name = simulation.name
    if "EdgePort" in port_names and isinstance(
        solution,
        (
            AnsysEigenmodeSolution,
            AnsysVoltageSolution,
            AnsysCurrentSolution,
        ),
    ):
        raise ValidateSimError(
            f"Simulation '{sim_name}' has at least one 'EdgePort'. This is incompatible with {type(solution)}",
            validation_type=has_edgeport_when_forbidden.__name__,
        )


def london_penetration_depth_with_ansys(simulation, solution):
    """Validation check: ensure that london penetration depth is not used with AnsysSolutions.
    Args:
        simulation: A Simulation object.
        solution: A Solution object.
    Raises:
        Errors when validation criteria are not met.
    """
    if not isinstance(solution, AnsysSolution):
        return

    material_dict = simulation.get_material_dict()
    for material, material_def in material_dict.items():
        if "london_penetration_depth" in material_def:
            raise ValidateSimError(
                f"Material {material} of simulation '{simulation.name}' defines London penetration "
                "depth, but the feature is not supported with Ansys solution."
            )


def flux_integration_layer_exists_if_needed(simulation, solution):
    """Validation check related to the presence of layers and magnetic flux integration.
    Args:
        simulation: A Simulation object.
        solution: A Solution object.
    Raises:
        Errors when validation criteria are not met.
    """
    sim_name = simulation.name
    has_integrate_flux = hasattr(solution, "integrate_magnetic_flux")
    integrate_flux = solution.integrate_magnetic_flux if has_integrate_flux else False

    # Ensures that a non-metal layer with thickness == 0 exists in the setup when "integrate_magnetic_flux" is True.
    if integrate_flux:
        layers = simulation.layers
        has_flux_integration_layer = False
        for layer in layers.values():
            if layer.get("thickness") == 0 and not simulation.is_metal(layer.get("material")):
                has_flux_integration_layer = True
                break
        if not has_flux_integration_layer:
            raise ValidateSimError(
                f"Simulation '{sim_name}' has 'integrate_magnetic_flux = True' "
                + "but the integration layer is missing.",
                validation_type=flux_integration_layer_exists_if_needed.__name__,
            )


def check_partition_region_naming(simulation, solution):
    """Validation check: Ensures that partition region names do not end with each other.

    Also checks that region names don't contain any of the common EPR groups.
    """

    integrate_energies = isinstance(solution, ElmerEPR3DSolution) or getattr(solution, "integrate_energies", False)

    if integrate_energies:
        common_groups = ["ma", "ms", "sa", "substrate", "vacuum"]
        regions = [p.name.lower() for p in simulation.get_partition_regions()]
        for i, r1 in enumerate(regions):
            for g in common_groups:
                if g in r1:
                    raise ValidateSimError(
                        f'Partition region names can not contain the common EPR group names ("{g}" in "{r1}")',
                        validation_type=check_partition_region_naming.__name__,
                    )
            for j, r2 in enumerate(regions):
                if i != j and r1.endswith(r2):
                    raise ValidateSimError(
                        f'Partition region names can not be suffixes of each other ("{r1}" ends with "{r2}")',
                        validation_type=check_partition_region_naming.__name__,
                    )


def check_tls_sheet_generation(simulation):
    """Validation check: Ensures that TLS sheets are correctly generated and not overlapping with sheet metals"""
    if simulation.tls_sheet_approximation:
        if simulation.detach_tls_sheets_from_body:
            if not recursive_all(simulation.tls_layer_thickness, lambda x: x > 0):
                raise ValidateSimError(
                    "Can't detach tls sheets from body if no positive `tls_layer_thickness` is given",
                    validation_type=check_tls_sheet_generation.__name__,
                )
        elif not recursive_all(simulation.metal_height, lambda x: x > 0):  # not detach_tls_sheets_from_body
            raise ValidateSimError(
                "TLS sheets can't be generated to overlap with sheet metals. Either use `metal_height > 0` "
                + "or `detach_tls_sheets_from_body=True`",
                validation_type=check_tls_sheet_generation.__name__,
            )


def check_tls_sheets_by_solution(simulation, solution):
    """Validation check: enforces solution type specific restrictions on TLS sheets"""
    if isinstance(solution, ElmerEPR3DSolution):
        # non-zero metal thickness and TLS sheets on metal boundary
        if simulation.tls_sheet_approximation and simulation.detach_tls_sheets_from_body:
            raise ValidateSimError(
                "ElmerEPR3DSolution requires simulation parameter `detach_tls_sheets_from_body` to be set False",
                validation_type=check_tls_sheets_by_solution.__name__,
            )

        m = simulation.metal_height
        if not recursive_all(m, lambda x: x > 0):
            raise ValidateSimError(
                f"ElmerEPR3DSolution requires non-zero `metal_height` to be used (found: {m})",
                validation_type=check_tls_sheets_by_solution.__name__,
            )
    elif isinstance(solution, (ElmerCapacitanceSolution, ElmerVectorHelmholtzSolution)):
        # TLS sheets detached from metal
        if simulation.tls_sheet_approximation and not simulation.detach_tls_sheets_from_body:
            raise ValidateSimError(
                "ElmerCapacitanceSolution and ElmerVectorHelmholtzSolution require simulation parameter "
                + "`detach_tls_sheets_from_body` to be set True",
                validation_type=check_tls_sheets_by_solution.__name__,
            )
    elif isinstance(solution, AnsysSolution):
        # TLS sheets detached from metal
        if simulation.tls_sheet_approximation and not simulation.detach_tls_sheets_from_body:
            logging.warning("By convention `detach_tls_sheets_from_body` should be True in Ansys simulations")


# Following code is not validation checks but utilities used by validity checks


def recursive_all(l, condition):
    if not isinstance(l, list):
        return condition(l)
    else:
        return all((recursive_all(e, condition) for e in l))


def get_port_names(simulation):
    """Helper function that returns a list of port names in a Simulation object.
    Args:
        simulation: A Simulation object.
    Returns:
        port_names: A list of names related to the ports present in simulation.
    """
    port_list = simulation.ports if isinstance(simulation, Simulation) else []
    port_names = []
    for port in port_list:
        port_names.append(type(port).__name__)
    return port_names


class ValidateSimError(Exception):
    """Custom exception class for specific error handling."""

    def __init__(self, message, validation_type=None):
        super().__init__(message)
        self.validation_type = validation_type



# This code is part of KQCircuits
# Copyright (C) 2021 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).
import logging
from os import cpu_count
from time import perf_counter

from kqcircuits.pya_resolver import pya
from kqcircuits.defaults import default_faces


class AreaReceiver(pya.TileOutputReceiver):
    """Class for handling and storing output from :class:`TilingProcessor`"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.area = 0.0

    def put(self, ix, iy, tile, obj, dbu, clip):
        """Function called by :class:`TilingProcessor` on output"""
        # pylint: disable=unused-argument
        logging.debug(f"Area for tile {ix},{iy}: {obj} ({dbu})")
        self.area = obj * (dbu * dbu)  # report as um^2


def get_area_and_density(cell: pya.Cell, layer_infos=None, optimize_ground_grid_calculations=True):
    """Get total area and density :math:`\\rho=\\frac{area}{bbox.area}` of all layers.

    This calculation is slow for geometries with many polygons, and in practice the layers containing ground grid
    take the majority of time. When ``optimize_ground_grid_calculations`` is set to ``True``, the ``ground_grid``
    layer area is calculated by assuming all polygons in this layer are identical and don't overlap each other
    (which they should be by definition). Further, the area of ``base_metal_gap`` is calculated by combining the areas
    of ``base_metal_gap_wo_grid`` and ``ground_grid`` areas.

    Args:
        cell: target cell to get area from
        layer_infos: list of ``LayerInfo`` to get area for, or None to get area for all layers.
        optimize_ground_grid_calculations:  ``True`` (default) to optimize ground grid area calculations.

    Returns: dictionary ``{layer_name: {'area': area, 'density': density}}``, where ``area`` is in um^2 and ``density``
       is a fraction < 1.
    """
    start_time = perf_counter()
    layout = cell.layout()

    all_layer_infos = {layer_info.name: layer_info for layer_info in layout.layer_infos()}
    if layer_infos is None:
        layer_infos = all_layer_infos.values()
    layer_infos = set(layer_infos)

    def _grid_area_and_density(layer_info):
        """Calculate the area and density for a layer where all shapes are known to be identical"""
        shapes = list(cell.begin_shapes_rec(layout.layer(layer_info)).each())
        shape_count = len(shapes)
        shape_area = shapes[0].shape().area() if shape_count > 0 else 0
        area = shape_count * float(shape_area)
        bbox_area = cell.bbox_per_layer(layout.layer(layer_info)).area()
        density = area / bbox_area if bbox_area != 0.0 else 0.0
        return area * layout.dbu**2, density

    # Separate out `ground_grid` and `base_metal_gap` layers in `layer_infos` for optimization
    ground_grid_faces = set()
    if optimize_ground_grid_calculations:
        for face, layers in default_faces.items():
            if (
                "ground_grid" in layers
                and "base_metal_gap" in layers
                and "base_metal_gap_wo_grid" in layers
                and (layers["ground_grid"] in layer_infos or layers["base_metal_gap"] in layer_infos)
            ):
                ground_grid_faces.add(face)
        for face in ground_grid_faces:
            ground_grid_layer = default_faces[face]["ground_grid"]
            base_metal_gap_wo_grid_layer = default_faces[face]["base_metal_gap_wo_grid"]
            base_metal_gap_layer = default_faces[face]["base_metal_gap"]

            layer_infos -= {ground_grid_layer, base_metal_gap_layer}  # Skip brute force calculation for these layers
            layer_infos.add(
                base_metal_gap_wo_grid_layer
            )  # Ensure base_metal_gap_wo_grid is included in the calculation

    # Perform tiled area calculation for all other layers
    tp = pya.TilingProcessor()
    tp.threads = cpu_count()
    tp.tile_size = (2000, 2000)  # microns
    layer_areas = [AreaReceiver() for _ in layer_infos]
    layer_bboxes = [AreaReceiver() for _ in layer_infos]
    for layer_info, area_receiver, bbox_receiver in zip(layer_infos, layer_areas, layer_bboxes):
        name = f"_{layer_info.name}"  # if `name` starts with a number, tp.execute() fails, so we add an underscore
        area, bbox = name + "_area", name + "_bbox"
        tp.input(name, cell.begin_shapes_rec(layout.layer(layer_info)))
        tp.output(area, area_receiver)
        tp.output(bbox, bbox_receiver)
        tp.queue(f"_output({area}, {name}.area)")
        tp.queue(f"_output({bbox}, {name}.bbox.area)")
    tp.execute("Calculate polygon and bounding box area")

    areas = [area.area for area in layer_areas]
    bboxes = [bbox.area for bbox in layer_bboxes]
    results = {
        layer_info.name: {"area": area, "density": area / bbox if bbox != 0.0 else 0.0}
        for layer_info, area, bbox in zip(layer_infos, areas, bboxes)
    }

    # Add optimized ground grid calculation results
    for face in ground_grid_faces:
        ground_grid_layer = default_faces[face]["ground_grid"]
        base_metal_gap_wo_grid_layer = default_faces[face]["base_metal_gap_wo_grid"]
        base_metal_gap_layer = default_faces[face]["base_metal_gap"]

        # Calculate ground grid area assuming all shapes in ``ground_grid`` are identical
        ground_grid_area, ground_grid_density = _grid_area_and_density(ground_grid_layer)
        results[ground_grid_layer.name] = {"area": ground_grid_area, "density": ground_grid_density}

        # Calcualte base_metal_gap area assuming the ground grid does not overlap with base_metal_gap_wo_grid
        gap_area = results[base_metal_gap_wo_grid_layer.name]["area"] + ground_grid_area
        gap_bbox = cell.bbox_per_layer(layout.layer(base_metal_gap_layer)).area() * layout.dbu**2
        gap_density = gap_area / gap_bbox if gap_bbox != 0.0 else 0.0
        results[base_metal_gap_layer.name] = {"area": gap_area, "density": gap_density}

    if len(results) > 0:
        logging.info(f"Area calculation took {perf_counter() - start_time:.1f} seconds")

    return results



# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

import logging
from typing import Callable
from kqcircuits.elements.element import Element
from kqcircuits.simulations.simulation import Simulation
from kqcircuits.simulations.partition_region import PartitionRegion
from kqcircuits.pya_resolver import pya


EPRTarget = Simulation | Element


def in_gui(element: EPRTarget) -> bool:
    """``partition_regions`` and ``correction_cuts`` can be called for elements
    when KLayout is in GUI mode, in which case element is not of type Simulation.
    This check can be used if ``partition_regions`` and ``correction_cuts``
    implementations use ``Simulation`` specific attributes and methods like
    ``box`` or ``face_stack``. If True is returned, some alternative
    attributes need to be used.
    """
    return not isinstance(element, Simulation)


def get_mer_z(element: EPRTarget, face_base: str) -> float:
    """Determines z position of a MER box postitioned at the base of the face
    given ``face_base`` as face id string."""
    if in_gui(element):
        return 0
    z_levels = element.face_z_levels()
    if face_base not in z_levels:
        logging.warning(
            f"Couldn't get z coordinate of face {face_base}, "
            "please check how 'face_stack' is defined. Defaulting to 0 for MER box."
        )
        return 0
    return z_levels[face_base][0]


def is_flip_chip(element: EPRTarget):
    """Determines if the geometry consists of multiple substrate layers"""
    if in_gui(element):
        return True
    return len(element.face_stack) > int(element.lower_box_height > 0) + 1


def extract_child_simulation(
    simulation: EPRTarget,
    refpoint_prefix: str | None = None,
    parameter_remap_function: Callable[[EPRTarget, str], any] | None = None,
    needed_parameters: list[str] | None = None,
) -> Simulation:
    """Given a simulation object that builds multiple elements within it,
    extracts a "child simulation" which is a stub object that only contains
    minimal set of refpoints and parameters from which element specific
    parition regions and correction cuts can be derived.

    When defining parition regions and correction cuts for an element
    under ``kqcircuits.simulations.epr``, it is recommended to create
    a function that takes arguments ``(simulation, refpoint_prefix, parameter_remap_function)``
    that passes the arguments to ``extract_child_simulation`` and explicitly
    lists ``needed_parameters``. See ``kqcircuits.simulations.epr.example``.

    Args:
        simulation: Simulation object that contains needed child simulation
        refpoint_prefix: Child simulation refpoints usually have a common prefix
            in refpoint names, which identify that the refpoint belongs to that child.
        parameter_remap_function: If ``simulation`` defines parameters that overrides
            parameters of the needed child simulation, these parameters should be mapped
            for the partition region and correction cut extraction functions to work.
            For example, suppose the correction cut location of some qubit is defined
            relative to its ``a`` parameter. But when the ``simulation`` inserts
            the qubit, it overrides qubit's ``a`` parameter with ``simulation``s
            ``qubit_a`` parameter. ``parameter_remap_function`` should be such that
            ``parameter_remap_function(simulation, "a") = simulation.qubit_a``.
        needed_parameters: List of all parameters that are needed to derive
            parition regions and correction cuts for needed child simulation.

    Returns:
        A "simulation object" that only contains minimal set of refpoints and parameters
        from which element specific parition regions and correction cuts can be derived.
        The result is a stub simulation object, so it doesn't have well defined
        build function or any features that would be useful to export a simulation by itself.
    """
    child_simulation = Simulation(simulation.layout)
    child_simulation.refpoints = (
        {
            # Strip away prefix to make this refpoint your own
            ref_name[len(refpoint_prefix) :]: ref_point
            for ref_name, ref_point in simulation.refpoints.items()
            if ref_name.startswith(refpoint_prefix)
        }
        if refpoint_prefix
        else dict(simulation.refpoints)
    )
    # Add default simulation parameters that should always be included
    needed_parameters = set(
        needed_parameters
        + [
            "face_ids",
            "face_stack",
            "lower_box_height",
            "upper_box_height",
            "substrate_height",
            "metal_height",
            "dielectric_height",
            "chip_distance",
        ]
    )
    for parameter in needed_parameters:
        if parameter_remap_function:
            value = parameter_remap_function(simulation, parameter)
        else:
            value = getattr(simulation, parameter)
        setattr(child_simulation, parameter, value)
    return child_simulation


def create_bulk_and_mer_partition_regions(
    name: str = "part",
    face: str | None = None,
    metal_edge_dimensions: list[float | None] | float | None = None,
    region: pya.DBox | pya.DPolygon | list[pya.DBox] | list[pya.DPolygon] | pya.Region | None = None,
    vertical_dimensions: list[float | None] | float | None = None,
    visualise: bool = True,
    mer: bool = True,
    bulk: bool = True,
) -> list[PartitionRegion]:
    """
    Creates both bulk and mer partition regions for element.
    """
    result = []
    if mer:
        result.append(
            PartitionRegion(
                name=f"{name}mer",
                face=face,
                metal_edge_dimensions=metal_edge_dimensions,
                region=region,
                vertical_dimensions=vertical_dimensions,
                visualise=visualise,
            )
        )
    if bulk:
        result.append(
            PartitionRegion(
                name=f"{name}bulk",
                face=face,
                metal_edge_dimensions=None,
                region=region,
                vertical_dimensions=vertical_dimensions,
                visualise=visualise,
            )
        )
    return result



# This code is part of KQCircuits
# Copyright (C) 2024 IQM Finland Oy
#
# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program. If not, see
# https://www.gnu.org/licenses/gpl-3.0.html.
#
# The software distribution should follow IQM trademark policy for open-source software
# (meetiqm.com/iqm-open-source-trademark-policy). IQM welcomes contributions to the code.
# Please see our contribution agreements for individuals (meetiqm.com/iqm-individual-contributor-license-agreement)
# and organizations (meetiqm.com/iqm-organization-contributor-license-agreement).

from kqcircuits.pya_resolver import pya
from kqcircuits.elements.finger_capacitor_square import eval_a2, eval_b2
from kqcircuits.simulations.epr.util import EPRTarget, create_bulk_and_mer_partition_regions
from kqcircuits.simulations.partition_region import PartitionRegion

vertical_dimension = 1.0
metal_edge_dimension = 1.0
waveguide_margin = 10.0
waveguide_length_scale = 5


def partition_regions(simulation: EPRTarget, prefix: str = "") -> list[PartitionRegion]:
    """
    Returns partition regions. Doesn't work in general with non-symmetric spiral capacitors.
    """
    rr = simulation.finger_width / 2 + simulation.ground_gap
    rr /= simulation.layout.dbu

    # Use modified SmoothCapacitor geometry to obtain partition regions on fingers
    finger_gaps = pya.Region(simulation.cell.begin_shapes_rec(simulation.get_layer("ground_grid_avoidance"))).sized(
        (-simulation.margin - simulation.ground_gap - simulation.finger_width / 2) / simulation.layout.dbu, 5
    )

    ra = simulation.a / 2 + simulation.b + waveguide_margin
    port_a = simulation.refpoints["port_a"]
    dir_a = waveguide_length_scale * (simulation.refpoints["port_a_corner"] - port_a)
    cross_a = ra / dir_a.abs() * pya.DVector(dir_a.y, -dir_a.x)
    wg_region = pya.Region(
        pya.DPolygon(
            [
                port_a - cross_a,
                port_a - cross_a + dir_a,
                port_a + cross_a + dir_a,
                port_a + cross_a,
            ]
        ).to_itype(simulation.layout.dbu)
    )
    rb = eval_a2(simulation) / 2 + eval_b2(simulation) + waveguide_margin
    port_b = simulation.refpoints["port_b"]
    dir_b = waveguide_length_scale * (simulation.refpoints["port_b_corner"] - port_b)
    cross_b = rb / dir_b.abs() * pya.DVector(dir_b.y, -dir_b.x)
    wg_region += pya.Region(
        pya.DPolygon(
            [
                port_b - cross_b,
                port_b - cross_b + dir_b,
                port_b + cross_b + dir_b,
                port_b + cross_b,
            ]
        ).to_itype(simulation.layout.dbu)
    )

    result = create_bulk_and_mer_partition_regions(
        name=f"{prefix}fingerg",
        face=simulation.face_ids[0],
        metal_edge_dimensions=metal_edge_dimension,
        region=finger_gaps,
        vertical_dimensions=vertical_dimension,
        bulk=False,
        visualise=True,
    )

    result += [
        PartitionRegion(
            name=f"{prefix}waveguides",
            face=simulation.face_ids[0],
            region=wg_region,
            visualise=True,
        )
    ]
    result += create_bulk_and_mer_partition_regions(
        name=f"{prefix}groundg",
        face=simulation.face_ids[0],
        metal_edge_dimensions=metal_edge_dimension,
        vertical_dimensions=vertical_dimension,
        bulk=False,
        visualise=True,
    )

    return result


def correction_cuts(simulation: EPRTarget, prefix: str = "") -> dict[str, dict]:
    """
    Returns correction cuts. Doesn't work in general with non-symmetric spiral capacitors.
    """

    base_rf = simulation.refpoints["base"]

    gaps = pya.Region(simulation.cell.begin_shapes_rec(simulation.get_layer("base_metal_gap_wo_grid")))
    top_most = None
    center_most = None
    for polygon in gaps.each():
        d_poly = polygon.to_dtype(simulation.layout.dbu)
        points = [(e.p1 + e.p2) / 2 for e in d_poly.each_edge()] + list(d_poly.each_point_hull())
        for p in points:
            if top_most is None or p.y > top_most.y:
                top_most = p
            if center_most is None or p.distance(base_rf) < center_most.distance(base_rf):
                center_most = p

    center_dir = 2 * (center_most - base_rf)
    center_gap_len = center_dir.abs()
    center_dir /= center_gap_len
    half_cut_len = 25.0
    result = {
        f"{prefix}fingergmer": {
            "p1": center_most + simulation.finger_width * 0.5 * center_dir,
            "p2": center_most - center_dir * (center_gap_len + simulation.finger_width * 0.5),
        },
        f"{prefix}groundgmer": {
            "p1": top_most + pya.DPoint(0, half_cut_len),
            "p2": top_most - pya.DPoint(0, simulation.ground_gap + simulation.finger_width * 0.5),
        },
    }
    return result
